@@ -3,14 +3,7 @@
 //! # Warning
 //!
 //! Please review the code before use in a production environment.
-//!
-//! The code has only been tested with a single DHCP server, and has not gone
-//! through any fuzzing.
-//!
-//! ## Limitations
-//!
-//! * No support for rebinding
-//! * No support for renewing
+//! This code has been tested, but only with a single DHCP server.
 //!
 //! # Feature Flags
 //!
@@ -36,10 +29,12 @@ mod pkt;
 mod rand;
 
 use hl::{
+    io::Seek,
     ll::{
-        net::{Eui48Addr, Ipv4Addr, SocketAddrV4},
+        net::{Eui48Addr, Ipv4Addr},
         LinkStatus, PhyCfg, Registers, Sn, SocketInterrupt, SocketInterruptMask,
     },
+    net::SocketAddrV4,
     Common, Error, Udp, UdpReader,
 };
 pub use w5500_hl as hl;
@@ -49,46 +44,31 @@ use pkt::{send_dhcp_discover, send_dhcp_request, MsgType, PktDe};
 pub use w5500_hl::Hostname;
 
 /// DHCP destination port.
-#[cfg(target_os = "none")]
 pub const DST_PORT: u16 = 67;
-/// DHCP destination port for testing on `std` targets.
-///
-/// When compiled for targets with `cfg(target_os = "none")` this is port 67.
-#[cfg(not(target_os = "none"))]
-pub const DST_PORT: u16 = 2050;
 
 /// DHCP source port.
-#[cfg(target_os = "none")]
 pub const SRC_PORT: u16 = 68;
-/// DHCP source port for testing on `std` targets.
-///
-/// When compiled for targets with `cfg(target_os = "none")` this is port 68.
-#[cfg(not(target_os = "none"))]
-pub const SRC_PORT: u16 = 2051;
-
-#[cfg(target_os = "none")]
-const DHCP_BROADCAST: SocketAddrV4 = SocketAddrV4::new(Ipv4Addr::BROADCAST, DST_PORT);
-#[cfg(not(target_os = "none"))]
-const DHCP_BROADCAST: SocketAddrV4 = SocketAddrV4::new(Ipv4Addr::LOCALHOST, DST_PORT);
-
-/// Duration in seconds to wait for the DHCP server to send a response.
-const TIMEOUT_SECS: u32 = 10;
 
 /// Duration in seconds to wait for physical link-up.
-const LINK_UP_TIMEOUT_SECS: u32 = 2;
+const LINK_UP_TIMEOUT_SECS: u32 = 1;
 
 /// DHCP client states.
 #[derive(Debug, PartialEq, Eq, Clone, Copy)]
 #[cfg_attr(feature = "defmt", derive(defmt::Format))]
-#[allow(dead_code)]
-enum State {
+#[non_exhaustive] // may support rebooting and ini-reboot in the future
+#[doc(hidden)]
+pub enum State {
+    /// Initialization state, client sends DHCPDISCOVER.
     Init,
+    /// Client waits for DHCPOFFER.
     Selecting,
+    /// Client sends for DHCPREQUEST.
     Requesting,
-    InitReboot,
-    Rebooting,
+    /// Client has a valid lease.
     Bound,
+    /// T1 expires, client sends DHCPREQUEST to renew.
     Renewing,
+    /// T2 expires, client sends DHCPREQUEST to rebind.
     Rebinding,
 }
 
@@ -114,11 +94,7 @@ enum State {
 /// // locally administered MAC address
 /// const MAC_ADDRESS: Eui48Addr = Eui48Addr::new(0x02, 0x00, 0x11, 0x22, 0x33, 0x44);
 ///
-/// // unwrap is not `const` yet
-/// const HOSTNAME: Hostname = match Hostname::new("example") {
-///     Some(hn) => hn,
-///     None => panic!("invalid hostname"),
-/// };
+/// const HOSTNAME: Hostname = Hostname::new_unwrapped("example");
 ///
 /// this_is_where_you_setup_the_w5500_int_pin_for_a_falling_edge_trigger();
 ///
@@ -130,7 +106,7 @@ enum State {
 ///
 /// let call_after_secs: u32 = dhcp.process(&mut w5500, monotonic_seconds())?;
 /// // call process again after call_after_secs, or on the next interrupt
-/// # Ok::<(), w5500_hl::Error<std::io::Error>>(())
+/// # Ok::<(), w5500_hl::Error<std::io::ErrorKind>>(())
 /// ```
 #[derive(Debug)]
 #[cfg_attr(feature = "defmt", derive(defmt::Format))]
@@ -139,8 +115,10 @@ pub struct Client<'a> {
     sn: Sn,
     /// DHCP client state
     state: State,
-    /// Timeout for DHCP server responses
-    timeout: Option<u32>,
+    /// Instant of the last state transition
+    state_timeout: Option<u32>,
+    /// Timeout duration in seconds
+    timeout: u32,
     /// Renewal duration.
     t1: u32,
     /// Rebinding duration.
@@ -163,6 +141,10 @@ pub struct Client<'a> {
     dns: Option<Ipv4Addr>,
     /// (S)NTP server
     ntp: Option<Ipv4Addr>,
+    /// Broadcast address
+    broadcast_addr: SocketAddrV4,
+    /// Source Port
+    src_port: u16,
 }
 
 impl<'a> Client<'a> {
@@ -183,11 +165,7 @@ impl<'a> Client<'a> {
     /// const DHCP_SN: Sn = Sn::Sn0;
     /// // locally administered MAC address
     /// const MAC_ADDRESS: Eui48Addr = Eui48Addr::new(0x02, 0x00, 0x11, 0x22, 0x33, 0x44);
-    /// // unwrap is not `const` yet
-    /// const HOSTNAME: Hostname = match Hostname::new("example") {
-    ///     Some(hostname) => hostname,
-    ///     None => panic!("invalid hostname"),
-    /// };
+    /// const HOSTNAME: Hostname = Hostname::new_unwrapped("example");
     /// let seed: u64 = rng.next_u64();
     ///
     /// let dhcp: Client = Client::new(DHCP_SN, seed, MAC_ADDRESS, HOSTNAME);
@@ -198,7 +176,8 @@ impl<'a> Client<'a> {
         Self {
             sn,
             state: State::Init,
-            timeout: None,
+            state_timeout: None,
+            timeout: 5,
             t1: 0,
             t2: 0,
             lease: 0,
@@ -210,10 +189,42 @@ impl<'a> Client<'a> {
             hostname,
             dns: None,
             ntp: None,
+            broadcast_addr: SocketAddrV4::new(Ipv4Addr::BROADCAST, DST_PORT),
+            src_port: SRC_PORT,
         }
     }
 
-    /// Returns `true` if the DHCP client is bound.
+    /// Set the DHCP state timeout duration in seconds.
+    ///
+    /// This is the duration to wait for the DHCP server to send a reply before
+    /// resetting and starting over.
+    ///
+    /// # Example
+    ///
+    /// Set a 10 second timeout.
+    ///
+    /// ```
+    /// use rand_core::RngCore;
+    /// use w5500_dhcp::{
+    ///     ll::{net::Eui48Addr, Sn},
+    ///     Client, Hostname,
+    /// };
+    /// # let mut rng = rand_core::OsRng;
+    ///
+    /// const HOSTNAME: Hostname = Hostname::new_unwrapped("example");
+    /// let mut dhcp: Client = Client::new(
+    ///     Sn::Sn0,
+    ///     rng.next_u64(),
+    ///     Eui48Addr::new(0x02, 0x00, 0x11, 0x22, 0x33, 0x44),
+    ///     HOSTNAME,
+    /// );
+    /// dhcp.set_timeout_secs(10);
+    /// ```
+    pub fn set_timeout_secs(&mut self, secs: u32) {
+        self.timeout = secs;
+    }
+
+    /// Returns `true` if the DHCP client has a valid lease.
     ///
     /// # Example
     ///
@@ -225,17 +236,21 @@ impl<'a> Client<'a> {
     /// };
     /// # let mut rng = rand_core::OsRng;
     ///
+    /// const HOSTNAME: Hostname = Hostname::new_unwrapped("example");
     /// let dhcp: Client = Client::new(
     ///     Sn::Sn0,
     ///     rng.next_u64(),
     ///     Eui48Addr::new(0x02, 0x00, 0x11, 0x22, 0x33, 0x44),
-    ///     Hostname::new("demo").unwrap(),
+    ///     HOSTNAME,
     /// );
-    /// assert_eq!(dhcp.is_bound(), false);
+    /// assert_eq!(dhcp.has_lease(), false);
     /// ```
     #[inline]
-    pub fn is_bound(&self) -> bool {
-        self.state == State::Bound
+    pub fn has_lease(&self) -> bool {
+        matches!(
+            self.state,
+            State::Bound | State::Rebinding | State::Renewing
+        )
     }
 
     /// Setup the DHCP socket interrupts.
@@ -251,21 +266,49 @@ impl<'a> Client<'a> {
         w5500.set_sn_imr(self.sn, MASK)?;
         w5500.close(self.sn)?;
         w5500.set_sipr(&self.ip)?;
-        w5500.udp_bind(self.sn, SRC_PORT)
+        w5500.udp_bind(self.sn, self.src_port)
     }
 
     fn timeout_elapsed_secs(&self, monotonic_secs: u32) -> Option<u32> {
-        self.timeout.map(|to| monotonic_secs - to)
+        self.state_timeout.map(|to| monotonic_secs - to)
     }
 
     fn next_call(&self, monotonic_secs: u32) -> u32 {
         if let Some(timeout_elapsed_secs) = self.timeout_elapsed_secs(monotonic_secs) {
-            TIMEOUT_SECS.saturating_sub(timeout_elapsed_secs)
+            self.timeout.saturating_sub(timeout_elapsed_secs)
         } else {
-            self.lease
-                .saturating_sub(monotonic_secs.saturating_sub(self.lease_monotonic_secs))
+            let elapsed: u32 = monotonic_secs.saturating_sub(self.lease_monotonic_secs);
+            match self.state {
+                State::Bound => self.t1.saturating_sub(elapsed),
+                State::Renewing => self.t2.saturating_sub(elapsed),
+                // rebinding
+                _ => self.lease.saturating_sub(elapsed),
+            }
+        }
+    }
+
+    fn set_state_with_timeout(&mut self, state: State, monotonic_secs: u32) {
+        debug!(
+            "{:?} -> {:?} with timeout {}",
+            self.state, state, monotonic_secs
+        );
+        self.state = state;
+        self.state_timeout = Some(monotonic_secs);
+    }
+
+    fn set_state(&mut self, state: State) {
+        debug!("{:?} -> {:?} without timeout", self.state, state);
+        self.state = state;
+        self.state_timeout = None;
+    }
+
+    /// Get the leased IP address.
+    pub fn leased_ip(&self) -> Option<Ipv4Addr> {
+        if self.has_lease() {
+            Some(self.ip)
+        } else {
+            None
         }
-        .saturating_add(1)
     }
 
     /// Get the DNS server provided by DHCP.
@@ -347,6 +390,17 @@ impl<'a> Client<'a> {
                 reader.header().origin
             );
 
+            let stream_len: u16 = reader.stream_len();
+            let header_len: u16 = reader.header().len;
+            if header_len > stream_len {
+                // this is often recoverable
+                warn!(
+                    "packet may be truncated header={} > stream={}",
+                    header_len, stream_len
+                );
+                return Ok(None);
+            }
+
             let mut pkt: PktDe<W5500> = PktDe::from(reader);
             if pkt.is_bootreply()? {
                 debug!("packet is not a bootreply");
@@ -362,47 +416,22 @@ impl<'a> Client<'a> {
         }
 
         while let Some(mut pkt) = recv(w5500, self.sn, self.xid)? {
+            debug!("{:?}", self.state);
             match self.state {
                 State::Selecting => {
                     self.ip = pkt.yiaddr()?;
                     pkt.done()?;
-                    self.request(w5500, monotonic_secs)?;
+                    self.request(w5500)?;
+                    self.set_state_with_timeout(State::Requesting, monotonic_secs);
                 }
                 State::Requesting | State::Renewing | State::Rebinding => {
                     match pkt.msg_type()? {
                         Some(MsgType::Ack) => {
-                            let subnet_mask: Ipv4Addr = match pkt.subnet_mask()? {
-                                Some(x) => x,
-                                None => {
-                                    error!("subnet_mask option missing");
-                                    return Ok(self.next_call(monotonic_secs));
-                                }
-                            };
-                            info!("subnet_mask: {}", subnet_mask);
-                            let gateway: Ipv4Addr = match pkt.dhcp_server()? {
-                                Some(x) => x,
-                                None => {
-                                    error!("gateway option missing");
-                                    return Ok(self.next_call(monotonic_secs));
-                                }
-                            };
-                            info!("gateway: {}", gateway);
-                            let renewal_time: u32 = match pkt.renewal_time()? {
-                                Some(x) => x,
-                                None => {
-                                    error!("renewal_time option missing");
-                                    return Ok(self.next_call(monotonic_secs));
-                                }
-                            };
-                            info!("renewal_time: {}", renewal_time);
-                            let rebinding_time: u32 = match pkt.rebinding_time()? {
-                                Some(x) => x,
-                                None => {
-                                    error!("rebinding_time option missing");
-                                    return Ok(self.next_call(monotonic_secs));
-                                }
-                            };
-                            info!("rebinding_time: {}", rebinding_time);
+                            let subnet_mask: Option<Ipv4Addr> = pkt.subnet_mask()?;
+                            let gateway: Option<Ipv4Addr> = pkt.dhcp_server()?;
+                            let dns: Option<Ipv4Addr> = pkt.dns()?;
+                            let ntp: Option<Ipv4Addr> = pkt.ntp()?;
+
                             let lease_time: u32 = match pkt.lease_time()? {
                                 Some(x) => x,
                                 None => {
@@ -411,31 +440,66 @@ impl<'a> Client<'a> {
                                 }
                             };
                             info!("lease_time: {}", lease_time);
+                            let renewal_time: u32 = match pkt.renewal_time()? {
+                                Some(x) => x,
+                                None => lease_time / 2,
+                            };
+                            info!("renewal_time: {}", renewal_time);
+                            let rebinding_time: u32 = match pkt.rebinding_time()? {
+                                Some(x) => x,
+                                None => lease_time.saturating_mul(7) / 8,
+                            };
+                            info!("rebinding_time: {}", rebinding_time);
 
-                            if let Some(dns) = pkt.dns()? {
+                            // de-rate times by 12%
+                            self.t1 = renewal_time.saturating_sub(renewal_time / 8);
+                            self.t2 = rebinding_time.saturating_sub(rebinding_time / 8);
+                            self.lease = lease_time.saturating_sub(lease_time / 8);
+                            self.lease_monotonic_secs = monotonic_secs;
+
+                            pkt.done()?;
+
+                            match subnet_mask {
+                                Some(subnet_mask) => {
+                                    info!("subnet_mask: {}", subnet_mask);
+                                    w5500.set_subr(&subnet_mask)?;
+                                }
+                                None if self.state == State::Renewing => (),
+                                None => {
+                                    error!("subnet_mask option missing");
+                                    return Ok(self.next_call(monotonic_secs));
+                                }
+                            };
+
+                            match gateway {
+                                Some(gateway) => {
+                                    info!("gateway: {}", gateway);
+                                    w5500.set_gar(&gateway)?;
+                                }
+                                None if self.state == State::Renewing => (),
+                                None => {
+                                    error!("gateway option missing");
+                                    return Ok(self.next_call(monotonic_secs));
+                                }
+                            };
+
+                            // rebinding and renewal do not need to set a new IP
+                            if self.state == State::Requesting {
+                                info!("dhcp.ip: {}", self.ip);
+                                w5500.set_sipr(&self.ip)?;
+                            }
+
+                            if let Some(dns) = dns {
                                 info!("DNS: {}", dns);
                                 self.dns.replace(dns);
                             };
-                            if let Some(ntp) = pkt.ntp()? {
+
+                            if let Some(ntp) = ntp {
                                 info!("NTP: {}", ntp);
                                 self.ntp.replace(ntp);
                             };
 
-                            self.t1 = renewal_time;
-                            self.t2 = rebinding_time;
-                            // de-rate lease time by 12%
-                            self.lease = lease_time.saturating_sub(lease_time / 8);
-                            self.lease_monotonic_secs = monotonic_secs;
-
-                            info!("dhcp.ip: {}", self.ip);
-
-                            pkt.done()?;
-                            w5500.set_subr(&subnet_mask)?;
-                            w5500.set_sipr(&self.ip)?;
-                            w5500.set_gar(&gateway)?;
-
-                            self.state = State::Bound;
-                            self.timeout = None;
+                            self.set_state(State::Bound);
                         }
                         Some(MsgType::Nak) => {
                             info!("request was NAK'd");
@@ -460,7 +524,7 @@ impl<'a> Client<'a> {
         }
 
         if let Some(elapsed_secs) = self.timeout_elapsed_secs(monotonic_secs) {
-            if elapsed_secs > TIMEOUT_SECS {
+            if elapsed_secs > self.timeout {
                 info!(
                     "timeout waiting for state to transition from {:?}",
                     self.state
@@ -471,20 +535,24 @@ impl<'a> Client<'a> {
             match self.state {
                 State::Init => self.discover(w5500, monotonic_secs)?,
                 // states handled by IRQs and timeouts
-                State::Selecting | State::Requesting | State::Renewing | State::Rebinding => (),
-                // states we do not care about (yet)
-                State::InitReboot | State::Rebooting => (),
-                State::Bound => {
+                State::Selecting | State::Requesting => (),
+                State::Bound | State::Renewing | State::Rebinding => {
                     let elapsed: u32 = monotonic_secs.wrapping_sub(self.lease_monotonic_secs);
-                    if elapsed > self.t1 {
-                        warn!("t1 expired, taking no action");
-                    }
-                    if elapsed > self.t2 {
-                        warn!("t2 expired, taking no action");
-                    }
                     if elapsed > self.lease {
                         info!("lease expired");
                         self.discover(w5500, monotonic_secs)?;
+                    } else if elapsed > self.t2
+                        && matches!(self.state, State::Bound | State::Renewing)
+                    {
+                        info!("t2 expired");
+                        self.request(w5500)?;
+                        // no need for timeout, lease expiration will handle failures
+                        self.set_state(State::Rebinding);
+                    } else if elapsed > self.t1 && matches!(self.state, State::Bound) {
+                        info!("t1 expired");
+                        self.request(w5500)?;
+                        // no need for timeout, t2 expiration will handle failures
+                        self.set_state(State::Renewing);
                     }
                 }
             }
@@ -503,26 +571,98 @@ impl<'a> Client<'a> {
         debug!("sending DHCPDISCOVER xid={:08X}", self.xid);
 
         w5500.set_sipr(&self.ip)?;
-        w5500.udp_bind(self.sn, SRC_PORT)?;
+        w5500.udp_bind(self.sn, self.src_port)?;
 
-        send_dhcp_discover(w5500, self.sn, &self.mac, self.hostname, self.xid)?;
-        self.state = State::Selecting;
-        self.timeout = Some(monotonic_secs);
+        send_dhcp_discover(
+            w5500,
+            self.sn,
+            &self.mac,
+            self.hostname,
+            self.xid,
+            &self.broadcast_addr,
+        )?;
+        self.set_state_with_timeout(State::Selecting, monotonic_secs);
         Ok(())
     }
 
-    fn request<W5500: Registers>(
-        &mut self,
-        w5500: &mut W5500,
-        monotonic_secs: u32,
-    ) -> Result<(), Error<W5500::Error>> {
+    fn request<W5500: Registers>(&mut self, w5500: &mut W5500) -> Result<(), Error<W5500::Error>> {
         self.xid = self.rand.next_u32();
         debug!("sending DHCPREQUEST xid={:08X}", self.xid);
-
         send_dhcp_request(w5500, self.sn, &self.mac, &self.ip, self.hostname, self.xid)?;
-
-        self.state = State::Requesting;
-        self.timeout = Some(monotonic_secs);
         Ok(())
     }
+
+    /// Set the DHCP source port.
+    ///
+    /// Defaults to [`SRC_PORT`].
+    /// This is an interface for testing, typically the default is what you
+    /// want to use.
+    #[inline]
+    #[doc(hidden)]
+    pub fn set_src_port(&mut self, port: u16) {
+        self.src_port = port
+    }
+
+    /// Set the client broadcast address.
+    ///
+    /// Defaults to [`Ipv4Addr::BROADCAST`]:[`DST_PORT`].
+    /// This is an interface for testing, typically the default is what you
+    /// want to use.
+    #[inline]
+    #[doc(hidden)]
+    pub fn set_broadcast_addr(&mut self, addr: SocketAddrV4) {
+        self.broadcast_addr = addr;
+    }
+
+    /// DHCP client state.
+    #[inline]
+    #[doc(hidden)]
+    pub fn state(&self) -> State {
+        self.state
+    }
+
+    /// T1 time.
+    ///
+    /// This should be used only as a debug interface, and not to set timers.
+    /// DHCP timers are tracked internally by [`process`](Self::process).
+    ///
+    /// Returns `None` if the DHCP client does not have a valid lease.
+    #[inline]
+    #[doc(hidden)]
+    pub fn t1(&self) -> Option<u32> {
+        match self.state {
+            State::Init | State::Selecting | State::Requesting => None,
+            State::Bound | State::Renewing | State::Rebinding => Some(self.t1),
+        }
+    }
+
+    /// T2 time.
+    ///
+    /// This should be used only as a debug interface, and not to set timers.
+    /// DHCP timers are tracked internally by [`process`](Self::process).
+    ///
+    /// Returns `None` if the DHCP client does not have a valid lease.
+    #[inline]
+    #[doc(hidden)]
+    pub fn t2(&self) -> Option<u32> {
+        match self.state {
+            State::Init | State::Selecting | State::Requesting => None,
+            State::Bound | State::Renewing | State::Rebinding => Some(self.t2),
+        }
+    }
+
+    /// Lease time.
+    ///
+    /// This should be used only as a debug interface, and not to set timers.
+    /// DHCP timers are tracked internally by [`process`](Self::process).
+    ///
+    /// Returns `None` if the DHCP client does not have a valid lease.
+    #[inline]
+    #[doc(hidden)]
+    pub fn lease_time(&self) -> Option<u32> {
+        match self.state {
+            State::Init | State::Selecting | State::Requesting => None,
+            State::Bound | State::Renewing | State::Rebinding => Some(self.lease),
+        }
+    }
 }
