@@ -1,7 +1,8 @@
 use w5500_hl::{
+    io::{Read, Seek, SeekFrom, Write},
     ll::{Registers, Sn},
-    net::{Eui48Addr, Ipv4Addr},
-    Common, Error, Hostname, Read, Seek, SeekFrom, UdpReader, Writer,
+    net::{Eui48Addr, Ipv4Addr, SocketAddrV4},
+    Error, Hostname, Udp, UdpReader, UdpWriter,
 };
 
 /// DHCP options.
@@ -124,11 +125,11 @@ const HW_ADDR_LEN: u8 = 6;
 
 #[derive(Debug)]
 struct PktSer<'a, W: Registers> {
-    writer: Writer<'a, W>,
+    writer: UdpWriter<'a, W>,
 }
 
-impl<'a, W: Registers> From<Writer<'a, W>> for PktSer<'a, W> {
-    fn from(writer: Writer<'a, W>) -> Self {
+impl<'a, W: Registers> From<UdpWriter<'a, W>> for PktSer<'a, W> {
+    fn from(writer: UdpWriter<'a, W>) -> Self {
         Self { writer }
     }
 }
@@ -318,7 +319,7 @@ impl<'a, W: Registers> PktSer<'a, W> {
         mac: &Eui48Addr,
         hostname: Hostname,
         xid: u32,
-    ) -> Result<Writer<'a, W>, Error<W::Error>> {
+    ) -> Result<UdpWriter<'a, W>, Error<W::Error>> {
         self.prepare_message(mac, xid)?;
         self.set_option_msg_type(MsgType::Discover)?;
         self.set_option_client_id(mac)?;
@@ -334,7 +335,7 @@ impl<'a, W: Registers> PktSer<'a, W> {
         ip: &Ipv4Addr,
         hostname: Hostname,
         xid: u32,
-    ) -> Result<Writer<'a, W>, Error<W::Error>> {
+    ) -> Result<UdpWriter<'a, W>, Error<W::Error>> {
         self.prepare_message(mac, xid)?;
         self.set_option_msg_type(MsgType::Request)?;
         self.set_option_client_id(mac)?;
@@ -352,11 +353,12 @@ pub fn send_dhcp_discover<W: Registers>(
     mac: &Eui48Addr,
     hostname: Hostname,
     xid: u32,
+    broadcast_addr: &SocketAddrV4,
 ) -> Result<(), Error<W::Error>> {
-    let writer: Writer<W> = w5500.writer(sn)?;
+    let writer: UdpWriter<W> = w5500.udp_writer(sn)?;
     PktSer::from(writer)
         .dhcp_discover(mac, hostname, xid)?
-        .udp_send_to(&crate::DHCP_BROADCAST)?;
+        .udp_send_to(broadcast_addr)?;
     Ok(())
 }
 
@@ -368,21 +370,40 @@ pub fn send_dhcp_request<W: Registers>(
     hostname: Hostname,
     xid: u32,
 ) -> Result<(), Error<W::Error>> {
-    let writer: Writer<W> = w5500.writer(sn)?;
+    let writer: UdpWriter<W> = w5500.udp_writer(sn)?;
     PktSer::from(writer)
         .dhcp_request(mac, ip, hostname, xid)?
         .send()?;
     Ok(())
 }
 
+#[derive(Debug)]
+struct OptionLocation {
+    option: u8,
+    ptr: u16,
+    len: u8,
+}
+
+impl OptionLocation {
+    const DEFAULT: Self = OptionLocation {
+        option: 0,
+        ptr: 0,
+        len: 0,
+    };
+}
+
 #[derive(Debug)]
 pub struct PktDe<'a, W: Registers> {
     reader: UdpReader<'a, W>,
+    option_location_cache: [OptionLocation; 9],
 }
 
 impl<'a, W: Registers> From<UdpReader<'a, W>> for PktDe<'a, W> {
     fn from(reader: UdpReader<'a, W>) -> Self {
-        Self { reader }
+        Self {
+            reader,
+            option_location_cache: [OptionLocation::DEFAULT; 9],
+        }
     }
 }
 
@@ -412,10 +433,43 @@ impl<'a, W: Registers> PktDe<'a, W> {
         Ok(buf.into())
     }
 
+    fn option_location_cache_insert(&mut self, option: u8, len: u8) {
+        if let Some(empty_entry) = self
+            .option_location_cache
+            .iter_mut()
+            .find(|l| l.option == 0)
+        {
+            *empty_entry = OptionLocation {
+                option,
+                ptr: self.reader.stream_position(),
+                len,
+            };
+        }
+    }
+
     /// Seeks to an option and returns the length if it exists.
     fn seek_to_option(&mut self, option: Options) -> Result<Option<u8>, Error<W::Error>> {
         let option: u8 = option.into();
-        self.reader.seek(SeekFrom::Start(240))?;
+
+        // check cache
+        let mut start_of_unknown_options: u16 = 240;
+        for location in self.option_location_cache.iter() {
+            if location.option != 0 {
+                if location.option == option {
+                    self.reader.seek(SeekFrom::Start(location.ptr))?;
+                    return Ok(Some(location.len));
+                }
+
+                start_of_unknown_options = location.ptr + u16::from(location.len);
+            } else {
+                break;
+            }
+        }
+
+        // seek to start of options
+        self.reader
+            .seek(SeekFrom::Start(start_of_unknown_options))?;
+
         loop {
             let (byte, len): (u8, u8) = {
                 let mut buf: [u8; 2] = [0; 2];
@@ -436,8 +490,10 @@ impl<'a, W: Registers> PktDe<'a, W> {
                     self.reader.seek(SeekFrom::Current(-1))?;
                 }
             } else if byte == option {
+                self.option_location_cache_insert(byte, len);
                 return Ok(Some(len));
             } else {
+                self.option_location_cache_insert(byte, len);
                 self.reader.seek(SeekFrom::Current(len.into()))?;
             }
         }
@@ -551,6 +607,7 @@ impl<'a, W: Registers> PktDe<'a, W> {
     }
 
     pub fn done(self) -> Result<(), W::Error> {
-        self.reader.done()
+        self.reader.done()?;
+        Ok(())
     }
 }
